#!/usr/bin/env python

## compressed sensing on noisy communication channels

import sys
import os
import math
import numpy as np
import cvxpy as cp
from math import sqrt
from amplpy import AMPL, DataFrame
import time
import bitarray
import matplotlib.pyplot as plt

################### configurable params #####################

myZero = 1e-9

# channel noise
DefaultDelta = 0.08

# encoding redundancy
DefaultRedundancy = 4

# distribution
DefaultDistribution = "normal"

# random projection
DefaultRP = True
jlleps = 0.2

#ShowPlots = True
ShowPlots = False

######################### functions #########################

# generate a componentwise Normal(0,1) matrix
def normalmatrix(m, n):
    return np.random.normal([0],[[1]],(m,n))

# generate a componentwise Uniform(0,1) matrix
def uniformmatrix01(m, n):
    return np.random.rand(m,n)

# generate a componentwise Uniform(-1,1) matrix
def uniformmatrix(m, n):
    return 2*np.random.rand(m,n)-1

def AAIeval(AA, m):
    ## check orthogonality 
    nA = AA.shape[0]
    ell2err = np.linalg.norm(np.subtract(AA, np.eye(nA))) / nA**2
    ell2errdiag = np.linalg.norm(np.subtract(np.diag(AA),np.ones(nA))) / nA
    ell1err = np.linalg.norm(np.subtract(AA, np.eye(nA)),1) / nA**2
    ell1errdiag = np.linalg.norm(np.subtract(np.diag(AA),np.ones(nA)),1) / nA
    print "  AAIeval: avg ||AA - I||_1 = ", ell1err
    print "  AAIeval: avg ||AA - I||_2 = ", ell2err
    print "  AAIeval: avg ||diag(AA) - diag(I)||_1 =", ell1errdiag
    print "  AAIeval: avg ||diag(AA) - diag(I)||_2 = ", ell2errdiag
    print "  AAIeval: ell1err/sqrt(m)={0:.4g}, ell2err={1:.4g}".format(ell1err/sqrt(m), ell2err)
    print "  AAIeval: ell1errdiag/sqrt(m)={0:.4g}, ell2errdiag={1:.4g}".format(ell1errdiag / sqrt(m), ell2errdiag)
    return(ell1err, ell1errdiag, ell2err, ell2errdiag)

def supp(x):
    S = []
    try:
        m,n = x.shape
    except:
        m = 0
        n = len(x)
    if m == 0:
        for j in range(n):
            if abs(x[j]) > myZero:
                S.append(j)
    else:
        for i in range(m):
            for j in range(m):
                if abs(x[i,j]) > myZero:
                    S.append((i,j))
    return S

def hamming(a,b):
    dist = 0
    la = len(a)
    lb = len(b)
    for i in range(min(la,lb)):
        if a[i] != b[i]:
            dist += 1
    dist += max(la,lb) - min(la,lb)
    return dist

# encode a string into a list of bits
def string2bitlist(txt):
    txt = txt.encode('ascii', 'ignore')
    ba = bitarray.bitarray()
    ba.fromstring(txt)
    w = ba.tolist()
    w = [int(i) for i in w]
    return w

# decode a list of bits into a string
def bitlist2string(w):
    if not type(w) is list:
        w = list(w)
    ba = bitarray.bitarray(w)
    return ba.tobytes()

# cap vector entries to given interval
def cap(v, interval):
    L = interval[0]
    U = interval[1]
    n = len(v)
    assert(L<U)
    #v = [min(U, max(L, vj)) for vj in v]
    for j in range(n):
        v[j] = min(U, max(L, v[j]))
    return v

# find mxn A, nxd Q encoding matrices with Im(A) orthog Im(Q)
def findencoding(m,n,d):
    # method 1: find random rotation
    M = uniformmatrix(n,n)
    #M = (M + M.T)/2
    MM = np.dot(M.T,M)
    (E,U) = np.linalg.eigh(MM)
    U = U[:,np.random.permutation(n)]
    A = U[:,range(m)].T
    Q = U[:,range(n-d,n)]
    return (A,Q)

####################### MAIN #######################

## read command line
nargs = len(sys.argv)
if nargs < 2:
    print "syntax: " + sys.argv[0] + "'sentence' [noise redundancy distrib rp jlleps]"
    print "  'sentence': a sentence to encode and decode"
    print "  noise = channel noise (error vector density), default", DefaultDelta
    print "  encoding has length redundancy*|sentence|, default", DefaultRedundancy
    print "  distrib = encoder distribution (uniform, normal), default", DefaultDistribution
    print "  rp = 1 activates random projections, default",
    if DefaultRP:
        print "1"
    else:
        print "0"
    print "  jlleps is the JLL epsilon if rp=1, default", jlleps
    exit(1)

# start the clocks
wallclock0 = time.time()
#usersys0 = time.process_time()

# set up defaults
rpflag = DefaultRP
distrib = DefaultDistribution
Redundancy = DefaultRedundancy
Delta = DefaultDelta    

## read command line
s = sys.argv[1]
if nargs >= 7:
    jlleps = float(sys.argv[6])
if nargs >= 6:
    if sys.argv[5] == "1":
        rpflag = True
    else:
        rpflag = False
if nargs >= 5:
    distrib = sys.argv[4]
if nargs >= 4:
    Redundancy = float(sys.argv[3])
if nargs >= 3:
    Delta = float(sys.argv[2])
    
print "noise (Delta) =", Delta
print "redundancy =", Redundancy
print "distribution:", distrib
print "random projection",
if rpflag:
    print "on"
    print "jlleps =", jlleps
else:
    print "off"
print "sentence:", s

## compute the bitlist corresponding to the string
w = string2bitlist(s)

## generate A,Q such that Im(A) \bot Im(Q)
# size of the encoding matrix Q: n x d
d = len(w)
n = int(round(Redundancy * d))
# size of the matrix A: m x n
#  Im(A) orthogonal to Im(Q), i.e. dim( Im(A.T|Q) ) = n
m = n - d # we can also wing a smaller number here
(A,Q) = findencoding(m,n,d)
print "A is {0:d} x {1:d}, Q is {2:d} x {3:d}".format(m,n,n,d)

# evaluate A
As = A / math.sqrt(m) 
AA = np.dot(As.T,As)
# evaluate orthogonality
print "checking orthogonality of A"
errs = AAIeval(AA, m)

## encode and send the string
z = np.dot(Q,w)

# generate random noise with density Delta
xhat = np.zeros(n)
for j in range(n):
    if np.random.uniform(0,1) <= Delta:
        xhat[j] = np.random.randn()
        
# send z and receive the noisy version zbar
zbar = z + xhat

## decode the noisy string
b = np.dot(A,zbar)

## random projection
if rpflag:
    k = int(round((1.0/(jlleps**2))*math.log(m)))
    if k >= m:
        print "rnd proj increases constrs from", m, "to", k, "deactivating"
    else:
        print "rnd proj decreases constrs from", m, "to", k
        T = normalmatrix(k,m) / sqrt(k)
        A = np.dot(T,A)
        b = np.dot(T,b)
        m = k
        print "  checking orthogonality of projected A"
        errs = AAIeval(AA, m)

        
## formulate and solve LP
print "solving basis pursuit LP"
x = cp.Variable(n)
obj = cp.Minimize(cp.norm(x, 1))
constrs = [A*x - b == 0]
bp = cp.Problem(obj, constrs)
#result = bp.solve(solver = cp.MOSEK)
cplexopts = {}
cplexopts['lpmethod'] = 4 # select barrier alg
cplexopts['barrier.crossover'] = -1 # no crossover
result = bp.solve(solver = cp.CPLEX, verbose=True, cplex_params=cplexopts)
print "  solverresult reported", result
objfunval = bp.value
xstar = np.array(x.value)

S = supp(xhat)
T = supp(xstar)
decoderr1 = np.linalg.norm(np.subtract(xstar, xhat), 1)
decoderr2 = np.linalg.norm(np.subtract(xstar, xhat))
print "  ||xhat||_1={0:.4g}, optobjfunval={1:.4g}".format(np.linalg.norm(xhat,1), np.linalg.norm(xstar,1))
print "  |supp(xhat)|={0:d}, ||supp(xstar)|={1:d}".format(len(S), len(T))
print "  ||xstar - xhat||_1={0:.4g}, ||xstar - xhat||_2={1:.4g}".format(decoderr1, decoderr2)

# retrieve corresponding string
zstar = zbar - xstar
pseudoinvQ = np.dot(np.linalg.inv(np.dot(Q.T,Q)),Q.T)
wstar = cap(np.around(np.dot(pseudoinvQ,zstar)), (0,1))
sstar = bitlist2string(wstar)
hdist = hamming(s,sstar)
print "hamming_distance(sent,retr) =", hdist
print "sent:", s
print "retr:", sstar

wallclock1 = time.time()
#usersys1 = time.process_time()
print "wallclock CPU: {0:.2f}".format(wallclock1-wallclock0)
#print "user+sys CPU: ", usersys1-usersys0

